/* Optimized (Intel Intrinsics) implementation of SpoC-64 AEAD
   Written by:
   Kalikinkar Mandal <kmandal@uwaterloo.ca>/<kmandal@unb.ca>
   Date: August 21, 2020
   This implementation supports when message or ad lengths are a multiple of 8 bits.
*/

#ifndef SLISCP_LIGHT192_H
#define SLISCP_LIGHT192_H

#include<stdint.h>
#include<x86intrin.h>
#include<immintrin.h>

#define STATE_SIZE	8
#define SIMECK_ROUND	6
#define STEP_ROUND	18
#define PARAL_INST_BY4	2

/*
   *SC1_192: step constants, applied on X_0
   *SC2_192: step constants, applied on X_2
*/
static const uint8_t SC1_192[18]={0x8, 0xc, 0xa, 0x2f, 0x38, 0x24, 0x36, 0xd, 0x2b, 0x3e, 0x1, 0x21, 0x11, 0x39, 0x5, 0x27, 0x34, 0x2e};// Step constants (SC_{2i})
static const uint8_t SC2_192[18]={0x29, 0x1d, 0x33, 0x2a, 0x1f, 0x10, 0x18, 0x14, 0x1e, 0x31, 0x9, 0x2d, 0x1b, 0x16, 0x3d, 0x3, 0x2, 0x23};// Step constants (SC_{2i+1})
/*
   *RC1_192: round constants of simeck box applied on X_1
   *RC2_192: round constants of simeck box applied on X_3
*/
static const uint8_t RC1_192[18]={0x7, 0x4, 0x6, 0x25, 0x17, 0x1c, 0x12, 0x3b, 0x26, 0x15, 0x3f, 0x20, 0x30, 0x28, 0x3c, 0x22, 0x13, 0x1a};// Round constants (RC_{2i})
static const uint8_t RC2_192[18]={0x27, 0x34, 0x2e, 0x19, 0x35, 0xf, 0x8, 0xc, 0xa, 0x2f, 0x38, 0x24, 0x36, 0xd, 0x2b, 0x3e, 0x1, 0x21};// Round constants (RC_{2i+1})

typedef unsigned long long int u64;
typedef unsigned int u32;
typedef unsigned int u8;

#define MASK	_mm256_set_epi32(0x00ffffff, 0x00ffffff, 0x00ffffff, 0x00ffffff, 0x00ffffff, 0x00ffffff, 0x00ffffff, 0x00ffffff)

#define ROT5(x) ((_mm256_slli_epi32(x, 5) | _mm256_srli_epi32(x, 19))&MASK)
#define ROT1(x) ((_mm256_slli_epi32(x, 1) | _mm256_srli_epi32(x, 23))&MASK)
#define SWAPREG1(x) (_mm256_permutevar8x32_epi32(x, _mm256_set_epi32(7, 5, 3, 1, 6, 4, 2, 0)))
#define RC(t1, t2) (_mm256_set_epi32(0x00fffffe^t2, 0x00fffffe^t1, 0x00fffffe^t2, 0x00fffffe^t1, 0x00fffffe^t2, 0x00fffffe^t1, 0x00fffffe^t2,0x00fffffe^t1))
#define SC(t1, t2) (_mm256_set_epi32(0x00ffff00^t2, 0x00ffffff, 0x00ffff00^t1, 0x00ffffff, 0x00ffff00^t2, 0x00ffffff, 0x00ffff00^t1, 0x00ffffff))
#define SWAPBLK(x) (_mm256_permute4x64_epi64(x, _MM_SHUFFLE(2,3,0,1)))
#define SWAPREG2(x) (_mm256_permutevar8x32_epi32(x, _mm256_set_epi32(7, 6, 3, 2, 5, 4, 1, 0)));

#define ROAX(x, y, t1, t2)\
{\
__m256i x2tmp;\
x = x&MASK;\
x2tmp = x;\
x = (ROT5(x)&x)^ROT1(x)^RC(t1, t2)^y;\
y = x2tmp;\
}

#define PACK_SSb(x, y)\
{\
__m256i xtmp, ytmp;\
xtmp = SWAPREG1(x);\
ytmp = SWAPREG1(y);\
x = _mm256_permute2x128_si256(xtmp,ytmp,0x20);\
y = _mm256_permute2x128_si256(xtmp,ytmp, 0x31);\
}

#define UNPACK_SSb(x, y)\
{\
__m256i xtmp, ytmp;\
xtmp = _mm256_unpacklo_epi32(x, y);\
ytmp = _mm256_unpackhi_epi32(x, y);\
x = _mm256_permute2x128_si256(xtmp, ytmp,0x20);\
y = _mm256_permute2x128_si256(xtmp, ytmp, 0x31);\
}


#define PACK(x, y, z, w)\
{\
__m256i x2tmp, x3tmp;\
x2tmp = SWAPREG2(x);\
x3tmp = SWAPREG2(z);\
x = _mm256_permute2x128_si256(x2tmp,x3tmp,0x20);\
z = _mm256_permute2x128_si256(x2tmp,x3tmp, 0x31);\
x2tmp = SWAPREG2(y);\
x3tmp = SWAPREG2(w);\
y = _mm256_permute2x128_si256(x2tmp,x3tmp,0x20);\
w = _mm256_permute2x128_si256(x2tmp,x3tmp, 0x31);\
}

#define UNPACK(x,y,z,w)\
{\
__m256i x2tmp, x3tmp;\
x2tmp = _mm256_unpacklo_epi64(x, z);\
x3tmp = _mm256_unpackhi_epi64(x, z);\
x = _mm256_permute2x128_si256(x2tmp,x3tmp,0x20);\
z = _mm256_permute2x128_si256(x2tmp,x3tmp, 0x31);\
x2tmp = _mm256_unpacklo_epi64(y, w);\
x3tmp = _mm256_unpackhi_epi64(y, w);\
y = _mm256_permute2x128_si256(x2tmp,x3tmp,0x20);\
w = _mm256_permute2x128_si256(x2tmp,x3tmp, 0x31);\
}

#define _mm256_storeu2_m128i(/* __m128i* */ hiaddr, /* __m128i* */ loaddr, \
		                             /* __m256i */ a) \
    do { \
	            __m256i _a = (a); \
	            _mm_storeu_si128((loaddr), _mm256_castsi256_si128(_a)); \
	            _mm_storeu_si128((hiaddr), _mm256_extractf128_si256(_a, 0x1)); \
	        } while (0)

void sliscp_light192_perm_test( u32 *state );
void crypto_aead_encrypt( u32 *tag, u32 tlen, u8 *ciphertext, u8 *plaintext, u32 plen, u8 *ad, u32 adlen, u8 *nonce, u8 *key );
void crypto_aead_decrypt( u8 *plaintext,  u8 *ciphertext, u32 plen, u32 *tag, u32 tlen, u8 *ad, u32 adlen, u8 *nonce, u8 *key);
#endif
