/* Optimized (Intel Intrinsics) implementation of SpoC-64 AEAD
   Written by:
   Kalikinkar Mandal <kmandal@uwaterloo.ca>/<kmandal@unb.ca>
   Date: August 21, 2020
   This implementation supports when message or ad lengths are a multiple of 8 bits.
*/

#include<stdio.h>
#include<stdlib.h>
#include<time.h>
#include<math.h>

#include "spoc64.h"
#include "clock_cycle.h"

#define NUM_ITER        2000
#define NUM_TEST      	500


void print_state ( u32 *state )
{
        u8 i, j;
        
        for ( j = 0; j < 2*PARAL_INST_BY2; j++ )
        {
                for ( i = 0; i < STATE_SIZE; i++ )
                        printf("%X", state[i+j*STATE_SIZE]);
                printf("\n");
        }
        return;
}

int main()
{
	//srand(time(NULL));
        u8 num_parallel_inst;
        u32 *state;
        int i, j;
        u64 count_cc;

	u8 *plaintext, *ciphertext;
	u32 *tag;
	u8 *ad;
	u8 *key, *nonce;
	u32 plen, tlen, klen, adlen;


        num_parallel_inst = 2*PARAL_INST_BY2;
	adlen = 16; // Associated data length = adlen*8;
	plen = 8*8+3; // Message length = plen*8;
	tlen = 4; // 64 = 32*2 bits
	klen = 16; // 128 = 16*8 bits

	key = (u8 *)malloc(sizeof(u8)*klen*num_parallel_inst);
	nonce = (u8 *)malloc(sizeof(u8)*klen*num_parallel_inst);

	tag = (u8 *)malloc(sizeof(u8)*tlen*num_parallel_inst);
	ad = (u8 *)malloc(sizeof(u8)*adlen*num_parallel_inst);
	plaintext = (u8 *)malloc(sizeof(u8)*plen*num_parallel_inst);
	ciphertext = (u8 *)malloc(sizeof(u8)*plen*num_parallel_inst);
        
        
        state = (u32 *)malloc(sizeof(u32)*num_parallel_inst*STATE_SIZE);
       
	//Randomly initilizes messages to states
        for ( i = 0; i < num_parallel_inst*STATE_SIZE; i++ )
                *(state+i) = 0x0;
        
        //print_state(state);
        sliscp_light192_perm_test ( state );
        //print_state(state);

        //Assinging 128-bit keys, nonces and messages//
	for ( i = 0; i < num_parallel_inst*klen; i++ )
	{
                *(nonce+i) = i%klen;
                *(key+i) = i%klen;
	}
	for ( i = 0; i < num_parallel_inst*plen; i++ )
                *(plaintext+i) = i%plen;
	
	for ( i = 0; i < num_parallel_inst*adlen; i++ )
                *(ad+i) = i%adlen;

        for ( i = 0; i < NUM_ITER+1; i++ )
        {
                count_cc = start_rdtsc();
                for ( j = 0; j < NUM_TEST; j++ )
                        sliscp_light192_perm_test ( state );
                count_cc = end_rdtsc()-count_cc;
                printf("Cycles per byte for sliscp_light_r18 = %f\n", (double)(count_cc)/(double)(num_parallel_inst*32*NUM_TEST));
        }

	//===================================================================================================================
				// AEAD: Encryption and Decryption Module//
	//===================================================================================================================

	printf("Nonce and Key:\n");
	for ( i = 0; i < num_parallel_inst; i++ )
	{
		printf("Nonce:\n");
		for ( j = 0; j < klen; j++ )
			printf("%02X", nonce[klen*i+j]);
		printf("\n");
		printf("Key:\n");
		for ( j = 0; j < klen; j++ )
			printf("%02X", key[klen*i+j]);
		printf("\n");
	}

        count_cc = start_rdtsc();
	crypto_aead_encrypt( tag, tlen, ciphertext, plaintext, plen, ad, adlen, nonce, key);
	count_cc = end_rdtsc()-count_cc;

	printf("Original plaintext:\n");
	for ( i = 0; i < num_parallel_inst; i++ )
	{
		for ( j = 0; j < plen; j++ )
			printf("%02X", plaintext[i*plen+j]);
		printf("\n");
	}
	printf("Ciphertext:\n");
	for ( i = 0; i < num_parallel_inst; i++ )
	{
		for ( j = 0; j < plen; j++ )
			printf("%02X", ciphertext[i*plen+j]);
		printf("\n");
	}
	printf("Tag:\n");
	for ( i = 0; i < num_parallel_inst; i++ )
        {
                for ( j = 0; j < tlen/2; j++ )
                        printf("%08X", ((tag[i*tlen+2*j]<<8)|(tag[i*tlen+2*j+1]>>16)));
                printf("\n");
        }

	crypto_aead_decrypt( plaintext, ciphertext, plen, tag, tlen, ad, adlen, nonce, key);
        
	printf("Decrypted plaintext:\n");
	for ( i = 0; i < num_parallel_inst; i++ )
	{
		for ( j = 0; j < plen; j++ )
			printf("%02X", plaintext[i*plen+j]);
		printf("\n");
	}
	
	printf("Tag:\n");
	for ( i = 0; i < num_parallel_inst; i++ )
	{
                for ( j = 0; j < tlen/2; j++ )
                        printf("%08X", ((tag[i*tlen+2*j]<<8)|(tag[i*tlen+2*j+1]>>16)));
                printf("\n");
	}
	
	printf("Encryption speed = %f cpb\n", (double)(count_cc)/(double)(num_parallel_inst*plen*4));
	
        
        free(state);
	free(plaintext);
	free(ciphertext);
	free(tag);
	free(key);
	free(nonce);
	free(ad);

return(0);
}
