/* Optimized (Intel Intrinsics) implementation of SpoC-64 AEAD
   Written by:
   Kalikinkar Mandal <kmandal@uwaterloo.ca>/<kmandal@unb.ca>
   Date: August 21, 2020
   This implementation supports when message or ad lengths are a multiple of 8 bits.
*/
//#include<stdio.h>
#include "spoc64.h"

/*
   *sliscp_light192: 18-round sliscp-light permutation
   *state: input state, and output is stored in state again (inplace)
*/
void sliscp_light192_perm_test ( u32 *state )
{
        u8 i, j, k;
        u32 t1, t2;
        
        __m128i x[4*PARAL_INST_BY2];
        __m128i xtmp, ytmp;
        
        for ( k = 0; k < PARAL_INST_BY2; k++ )
                PACK(x[4*k], x[4*k+1], x[4*k+2], x[4*k+3], state, 16*k, 16*k+4, 16*k+8, 16*k+12);
        
        for ( i = 0; i < STEP_ROUND; i++ )
        {
                
                for ( k = 0; k < PARAL_INST_BY2; k++ )
                {
                        //SSb
                        PACK_SSb(x[4*k+2],x[4*k+3]);
                        for ( j = 0; j < SIMECK_ROUND; j++ )
                        {
                                t1 = (u32)((RC1_192[i] >> j)&1);
                                t2 = (u32)((RC2_192[i] >> j)&1);
                                ROAX(x[4*k+2], x[4*k+3], t1, t2);
                        }
                        UNPACK_SSb(x[4*k+2],x[4*k+3]);
                        //ASc
                        t1 = (u32)SC1_192[i];
                        t2 = (u32)SC2_192[i];
                        xtmp = x[4*k+0]^SC(t1, t2);
                        ytmp = x[4*k+1]^SC(t1, t2);
                        //MSb
                        x[4*k+0] = x[4*k+2];
                        x[4*k+1] = x[4*k+3];
                        x[4*k+2]^=xtmp;
                        x[4*k+3]^=ytmp;
                        x[4*k+2] = SWAPBLK(x[4*k+2]);
                        x[4*k+3] = SWAPBLK(x[4*k+3]);
                }
        }
        //Unpacking state
        for ( k = 0; k < PARAL_INST_BY2; k++)
        {
                xtmp = _mm_unpacklo_epi64(x[4*k+0], x[4*k+2]);
                ytmp = _mm_unpackhi_epi64(x[4*k+0], x[4*k+2]);
                _mm_storeu_si128((void *) (state + 0+16*k),xtmp);
                _mm_storeu_si128((void *) (state + 4 +16*k),ytmp);
                xtmp = _mm_unpacklo_epi64(x[4*k+1], x[4*k+3]);
                ytmp = _mm_unpackhi_epi64(x[4*k+1], x[4*k+3]);
                _mm_storeu_si128((void *) (state + 8+16*k),xtmp);
                _mm_storeu_si128((void *) (state + 12+16*k),ytmp);
        }
        return;
}

/*
   *sliscp_light192: 18-round sliscp-light permutation of width 192 bits
   *x: input state, and output state is stored in state (inplace)
*/

void sliscp_light192( __m128i *x )
{
	__m128i xtmp, ytmp;
	u8 i, j, k;
	u32 t1, t2;

	for ( i = 0; i < STEP_ROUND; i++ )
	{
		for ( k = 0; k < PARAL_INST_BY2; k++ )
		{
			//SSb
        		PACK_SSb(x[4*k+2], x[4*k+3]);
        		for ( j = 0; j < SIMECK_ROUND; j++ )
        		{
				t1 = (u32)((RC1_192[i] >> j)&1);
                		t2 = (u32)((RC2_192[i] >> j)&1);
                		ROAX( x[4*k+2], x[4*k+3], t1, t2);
        		}
        		UNPACK_SSb(x[4*k+2], x[4*k+3]);
        		//ASc
        		t1 = (u32)SC1_192[i];
        		t2 = (u32)SC2_192[i];
        		xtmp = x[4*k+0]^SC(t1, t2);
        		ytmp = x[4*k+1]^SC(t1, t2);
        		//MSb
			x[4*k+0] = x[4*k+2];
			x[4*k+1] = x[4*k+3];
        		x[4*k+2]^=xtmp;
        		x[4*k+3]^=ytmp;
        		x[4*k+2] = SWAPBLK(x[4*k+2]);
        		x[4*k+3] = SWAPBLK(x[4*k+3]);
		}
	}
}


/*
   *crypto_aead_encrypt: encrypt plaintext and produce tag
   *key: key 
   *nonce: public nonce
   *adlen: length of ad
   *ad: associate data
   *plen: length of plaintext
   *plaintext: plaintext to be encrypted
   *ciphertext: ciphertext
   *tlen: length of tag (in 32-bit word)
   *tag: tag
*/
void crypto_aead_encrypt( u32 *tag, u32 tlen, u8 *ciphertext, u8 *plaintext, u32 plen, u8 *ad, u32 adlen, u8 *nonce, u8 *key)
{
        u8 k, i;
        u32 l, mblock, adblock, lmblock, ladblock;
        __m128i x[4*PARAL_INST_BY2];
	u8 tmp2[8], tmp3[8];
	u32 index1, index2;
	mblock = plen/8;
        lmblock = plen%8;
	adblock = adlen/8;
        ladblock = adlen%8;
       
	//Initialization: Loading Key & Nonce

	for ( i = 0; i < PARAL_INST_BY2; i++ )
	{
		x[4*i] = _mm_set_epi8(0x0, nonce[32*i+7], key[32*i+14], key[32*i+15],0x0, nonce[32*i+4],nonce[32*i+5],nonce[32*i+6],0x0, nonce[32*i+3], key[32*i+6], key[32*i+7],0x0, nonce[32*i+0],nonce[32*i+1],nonce[32*i+2]);

		x[4*i+1] = _mm_set_epi8( 0x0, nonce[32*i+23],key[32*i+30],key[32*i+31],0x0, nonce[32*i+20],nonce[32*i+21],nonce[32*i+22], 0x0, nonce[32*i+19],key[32*i+22],key[32*i+23], 0x0, nonce[32*i+16],nonce[32*i+17],nonce[32*i+18]);

		x[4*i+2] = _mm_set_epi8(0x0,key[32*i+11],key[32*i+12],key[32*i+13],0x0, key[32*i+8],key[32*i+9],key[32*i+10], 0x0,key[32*i+3],key[32*i+4],key[32*i+5],0x0, key[32*i+0],key[32*i+1],key[32*i+2]);

		x[4*i+3] = _mm_set_epi8(0x0,key[32*i+27],key[32*i+28],key[32*i+29],0x0, key[32*i+24],key[32*i+25],key[32*i+26], 0x0,key[32*i+19],key[32*i+20],key[32*i+21],0x0,key[32*i+16],key[32*i+17],key[32*i+18]);
	}
	sliscp_light192(x);
        
	for ( i = 0; i < PARAL_INST_BY2; i++ )
	{
                x[4*i+2] = x[4*i+2]^_mm_set_epi8(0x0,nonce[32*i+15],0x0,0x0, 0x0,nonce[32*i+12], nonce[32*i+13], nonce[32*i+14], 0x0,nonce[32*i+11],0x0,0x0, 0x0, nonce[32*i+8], nonce[32*i+9], nonce[32*i+10]);
                x[4*i+3] = x[4*i+3]^_mm_set_epi8(0x0, nonce[32*i+31], 0x0,0x0,0x0,nonce[32*i+28], nonce[32*i+29], nonce[32*i+30], 0x0,nonce[32*i+27], 0x0, 0x0,0x0,nonce[32*i+24], nonce[32*i+25], nonce[32*i+26]);
	}
	//Proceesing Associated Data
	for ( l = 0; l < adblock; l++ )
	{
                sliscp_light192(x);
		for ( k = 0; k < PARAL_INST_BY2; k++)
		{
			index1 = 8*l + adlen*2*k;
			index2 = 8*l + adlen*2*k+adlen;
			x[4*k+2] = x[4*k+2]^_mm_set_epi8(0x0, ad[index1+7], 0x0, 0x0, 0x0, ad[index1+4], ad[index1+5], ad[index1+6], 0x0, ad[index1+3], 0x0, 0x0, 0x0, ad[index1], ad[index1+1], ad[index1+2]);
			x[4*k+3] = x[4*k+3]^_mm_set_epi8(0x0, ad[index2+7], 0x0, 0x0, 0x0, ad[index2+4], ad[index2+5], ad[index2+6], 0x0, ad[index2+3], 0x0, 0x0, 0x0, ad[index2], ad[index2+1], ad[index2+2]);
			
			//Domain seperator
			x[4*k] = x[4*k]^_mm_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x0, 0x00);
			x[4*k+1] = x[4*k+1]^_mm_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x0, 0x00);
		}
	}
	if ( ladblock != 0 )
	{
                sliscp_light192(x);
		for ( k = 0; k < PARAL_INST_BY2; k++)
		{
			index1 = 8*l + adlen*2*k;
			index2 = 8*l + adlen*2*k+adlen;
			for ( i = 0; i < ladblock; i++ )
			{
				tmp2[i] = ad[index1+i];
				tmp3[i] = ad[index2+i];
			}
			tmp2[ladblock] = 0x80;
			tmp3[ladblock] = 0x80;
			for ( i = ladblock+1; i < 8; i++ )
			{
				tmp2[i] = 0x0;
				tmp3[i] = 0x0;
			}
			x[4*k+2] = x[4*k+2]^_mm_set_epi8(0x0, tmp2[7], 0x0, 0x0, 0x0, tmp2[4], tmp2[5], tmp2[6], 0x0, tmp2[3], 0x0, 0x0, 0x0, tmp2[0], tmp2[1], tmp2[2]);
			x[4*k+3] = x[4*k+3]^_mm_set_epi8(0x0, tmp3[7], 0x0, 0x0, 0x0, tmp3[4], tmp3[5], tmp3[6], 0x0, tmp3[3], 0x0, 0x0, 0x0, tmp3[0], tmp3[1], tmp3[2]);
			//Domain seperator
			x[4*k] = x[4*k]^_mm_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x30, 0x0, 0x00);
			x[4*k+1] = x[4*k+1]^_mm_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x30, 0x0, 0x00);
		}
	}
	//Encrypting plaintext
	for ( l = 0; l < mblock; l++ )
	{
                sliscp_light192(x);
		for ( k = 0; k < PARAL_INST_BY2; k++)
		{
			index1 = 8*l + plen*2*k;
			index2 = 8*l + plen*2*k+plen;
			ciphertext[index1] = _mm_extract_epi8 (x[4*k], 2)^plaintext[index1];
			ciphertext[index1+1] = _mm_extract_epi8 (x[4*k], 1)^plaintext[index1+1];
			ciphertext[index1+2] = _mm_extract_epi8 (x[4*k], 0)^plaintext[index1+2];
			ciphertext[index1+3] = _mm_extract_epi8 (x[4*k], 6)^plaintext[index1+3];
			ciphertext[index1+4] = _mm_extract_epi8 (x[4*k], 10)^plaintext[index1+4];
			ciphertext[index1+5] = _mm_extract_epi8 (x[4*k], 9)^plaintext[index1+5];
			ciphertext[index1+6] = _mm_extract_epi8 (x[4*k], 8)^plaintext[index1+6];
			ciphertext[index1+7] = _mm_extract_epi8 (x[4*k], 14)^plaintext[index1+7];

			x[4*k+2] = x[4*k+2]^_mm_set_epi8(0x0, plaintext[index1+7], 0x0, 0x0, 0x0, plaintext[index1+4], plaintext[index1+5], plaintext[index1+6], 0x0, plaintext[index1+3], 0x0, 0x0, 0x0, plaintext[index1], plaintext[index1+1], plaintext[index1+2]);
			
			ciphertext[index2] = _mm_extract_epi8 (x[4*k+1], 2)^plaintext[index2];
			ciphertext[index2+1] = _mm_extract_epi8 (x[4*k+1], 1)^plaintext[index2+1];
			ciphertext[index2+2] = _mm_extract_epi8 (x[4*k+1], 0)^plaintext[index2+2];
			ciphertext[index2+3] = _mm_extract_epi8 (x[4*k+1], 6)^plaintext[index2+3];
			ciphertext[index2+4] = _mm_extract_epi8 (x[4*k+1], 10)^plaintext[index2+4];
			ciphertext[index2+5] = _mm_extract_epi8 (x[4*k+1], 9)^plaintext[index2+5];
			ciphertext[index2+6] = _mm_extract_epi8 (x[4*k+1], 8)^plaintext[index2+6];
			ciphertext[index2+7] = _mm_extract_epi8 (x[4*k+1], 14)^plaintext[index2+7];
			x[4*k+3] = x[4*k+3]^_mm_set_epi8(0x0, plaintext[index2+7], 0x0, 0x0, 0x0, plaintext[index2+4], plaintext[index2+5], plaintext[index2+6], 0x0, plaintext[index2+3], 0x0, 0x0, 0x0, plaintext[index2], plaintext[index2+1], plaintext[index2+2]);
			
			//Domain seperator
			x[4*k] = x[4*k]^_mm_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x0, 0x00);
			x[4*k+1] = x[4*k+1]^_mm_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x0, 0x00);
		}
	}
	if ( lmblock != 0 )
	{
                sliscp_light192(x);
		for ( k = 0; k < PARAL_INST_BY2; k++)
		{
			index1 = 8*l + plen*2*k;
			index2 = 8*l + plen*2*k+plen;
			if ( lmblock == 1 )
			{
				ciphertext[index1] = _mm_extract_epi8 (x[4*k], 2)^plaintext[index1];
				ciphertext[index2] = _mm_extract_epi8 (x[4*k+1], 2)^plaintext[index2];
			}
			else if ( lmblock == 2 )
			{
				ciphertext[index1] = _mm_extract_epi8 (x[4*k], 2)^plaintext[index1];
				ciphertext[index1+1] = _mm_extract_epi8 (x[4*k], 1)^plaintext[index1+1];
				ciphertext[index2] = _mm_extract_epi8 (x[4*k+1], 2)^plaintext[index2];
				ciphertext[index2+1] = _mm_extract_epi8 (x[4*k+1], 1)^plaintext[index2+1];
			}
			else if ( lmblock == 3 )
			{
				ciphertext[index1] = _mm_extract_epi8 (x[4*k], 2)^plaintext[index1];
				ciphertext[index1+1] = _mm_extract_epi8 (x[4*k], 1)^plaintext[index1+1];
				ciphertext[index1+2] = _mm_extract_epi8 (x[4*k], 0)^plaintext[index1+2];
				ciphertext[index2] = _mm_extract_epi8 (x[4*k+1], 2)^plaintext[index2];
				ciphertext[index2+1] = _mm_extract_epi8 (x[4*k+1], 1)^plaintext[index2+1];
				ciphertext[index2+2] = _mm_extract_epi8 (x[4*k+1], 0)^plaintext[index2+2];
			}
			else if ( lmblock == 4 )
			{
				ciphertext[index1] = _mm_extract_epi8 (x[4*k], 2)^plaintext[index1];
				ciphertext[index1+1] = _mm_extract_epi8 (x[4*k], 1)^plaintext[index1+1];
				ciphertext[index1+2] = _mm_extract_epi8 (x[4*k], 0)^plaintext[index1+2];
				ciphertext[index1+3] = _mm_extract_epi8 (x[4*k], 6)^plaintext[index1+3];
				ciphertext[index2] = _mm_extract_epi8 (x[4*k+1], 2)^plaintext[index2];
				ciphertext[index2+1] = _mm_extract_epi8 (x[4*k+1], 1)^plaintext[index2+1];
				ciphertext[index2+2] = _mm_extract_epi8 (x[4*k+1], 0)^plaintext[index2+2];
				ciphertext[index2+3] = _mm_extract_epi8 (x[4*k+1], 6)^plaintext[index2+3];
			}
			else if ( lmblock == 5 )
			{
				ciphertext[index1] = _mm_extract_epi8 (x[4*k], 2)^plaintext[index1];
				ciphertext[index1+1] = _mm_extract_epi8 (x[4*k], 1)^plaintext[index1+1];
				ciphertext[index1+2] = _mm_extract_epi8 (x[4*k], 0)^plaintext[index1+2];
				ciphertext[index1+3] = _mm_extract_epi8 (x[4*k], 6)^plaintext[index1+3];
				ciphertext[index1+4] = _mm_extract_epi8 (x[4*k], 10)^plaintext[index1+4];
				ciphertext[index2] = _mm_extract_epi8 (x[4*k+1], 2)^plaintext[index2];
				ciphertext[index2+1] = _mm_extract_epi8 (x[4*k+1], 1)^plaintext[index2+1];
				ciphertext[index2+2] = _mm_extract_epi8 (x[4*k+1], 0)^plaintext[index2+2];
				ciphertext[index2+3] = _mm_extract_epi8 (x[4*k+1], 6)^plaintext[index2+3];
				ciphertext[index2+4] = _mm_extract_epi8 (x[4*k+1], 10)^plaintext[index2+4];
			}
			else if ( lmblock == 6 )
			{
				ciphertext[index1] = _mm_extract_epi8 (x[4*k], 2)^plaintext[index1];
				ciphertext[index1+1] = _mm_extract_epi8 (x[4*k], 1)^plaintext[index1+1];
				ciphertext[index1+2] = _mm_extract_epi8 (x[4*k], 0)^plaintext[index1+2];
				ciphertext[index1+3] = _mm_extract_epi8 (x[4*k], 6)^plaintext[index1+3];
				ciphertext[index1+4] = _mm_extract_epi8 (x[4*k], 10)^plaintext[index1+4];
				ciphertext[index1+5] = _mm_extract_epi8 (x[4*k], 9)^plaintext[index1+5];
				ciphertext[index2] = _mm_extract_epi8 (x[4*k+1], 2)^plaintext[index2];
				ciphertext[index2+1] = _mm_extract_epi8 (x[4*k+1], 1)^plaintext[index2+1];
				ciphertext[index2+2] = _mm_extract_epi8 (x[4*k+1], 0)^plaintext[index2+2];
				ciphertext[index2+3] = _mm_extract_epi8 (x[4*k+1], 6)^plaintext[index2+3];
				ciphertext[index2+4] = _mm_extract_epi8 (x[4*k+1], 10)^plaintext[index2+4];
				ciphertext[index2+5] = _mm_extract_epi8 (x[4*k+1], 9)^plaintext[index2+5];
			}
			else if ( lmblock == 7 )
			{
				ciphertext[index1] = _mm_extract_epi8 (x[4*k], 2)^plaintext[index1];
				ciphertext[index1+1] = _mm_extract_epi8 (x[4*k], 1)^plaintext[index1+1];
				ciphertext[index1+2] = _mm_extract_epi8 (x[4*k], 0)^plaintext[index1+2];
				ciphertext[index1+3] = _mm_extract_epi8 (x[4*k], 6)^plaintext[index1+3];
				ciphertext[index1+4] = _mm_extract_epi8 (x[4*k], 10)^plaintext[index1+4];
				ciphertext[index1+5] = _mm_extract_epi8 (x[4*k], 9)^plaintext[index1+5];
				ciphertext[index1+6] = _mm_extract_epi8 (x[4*k], 8)^plaintext[index1+6];
				ciphertext[index2] = _mm_extract_epi8 (x[4*k+1], 2)^plaintext[index2];
				ciphertext[index2+1] = _mm_extract_epi8 (x[4*k+1], 1)^plaintext[index2+1];
				ciphertext[index2+2] = _mm_extract_epi8 (x[4*k+1], 0)^plaintext[index2+2];
				ciphertext[index2+3] = _mm_extract_epi8 (x[4*k+1], 6)^plaintext[index2+3];
				ciphertext[index2+4] = _mm_extract_epi8 (x[4*k+1], 10)^plaintext[index2+4];
				ciphertext[index2+5] = _mm_extract_epi8 (x[4*k+1], 9)^plaintext[index2+5];
				ciphertext[index2+6] = _mm_extract_epi8 (x[4*k+1], 8)^plaintext[index2+6];
			}
			for ( i = 0; i < lmblock; i++ )
			{
				tmp2[i] = plaintext[index1+i];
				tmp3[i] = plaintext[index2+i];
			}
			tmp2[lmblock] = 0x80;
			tmp3[lmblock] = 0x80;
			for ( i = lmblock+1; i < 8; i++ )
			{
				tmp2[i] = 0x0;
				tmp3[i] = 0x0;
			}
			x[4*k+2] = x[4*k+2]^_mm_set_epi8(0x0, tmp2[7], 0x0, 0x0, 0x0, tmp2[4], tmp2[5], tmp2[6], 0x0, tmp2[3], 0x0, 0x0, 0x0, tmp2[0], tmp2[1], tmp2[2]);
			x[4*k+3] = x[4*k+3]^_mm_set_epi8(0x0, tmp3[7], 0x0, 0x0, 0x0, tmp3[4], tmp3[5], tmp3[6], 0x0, tmp3[3], 0x0, 0x0, 0x0, tmp3[0], tmp3[1], tmp3[2]);
			//Domain seperator
			x[4*k] = x[4*k]^_mm_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x50, 0x0, 0x00);
			x[4*k+1] = x[4*k+1]^_mm_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x50, 0x0, 0x00);
		}
	}
        
	//Tag generation phase
	for ( k = 0; k < PARAL_INST_BY2; k++)
	{
        	//Domain seperator
		x[4*k] = x[4*k]^_mm_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x80, 0x0, 0x00);
		x[4*k+1] = x[4*k+1]^_mm_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x80, 0x0, 0x00);
	}
        sliscp_light192(x);
	//Extracting tag
	for ( k = 0; k < PARAL_INST_BY2; k++)
	{
		_mm_storeu_si128((void *) (tag + 0 + 8*k), x[4*k+2]);
		_mm_storeu_si128((void *) (tag + 4 + 8*k), x[4*k+3]);
	}
return;
}


/*
   *crypto_aead_decrypt: decrypt ciphertext
   *key: key 
   *nonce: public nonce
   *adlen: length of ad (in byte)
   *ad: associate data
   *tlen: length of tag (in 32-bit)
   *tag: tag
   *plen: length of plaintext (in byte word)
   *plaintext: plaintext to be encrypted
   *ciphertext: ciphertext
*/

void crypto_aead_decrypt( u8 *plaintext, u8 *ciphertext, u32 plen, u32 *tag, u32 tlen, u8 *ad, u32 adlen, u8 *nonce, u8 *key)
{
        u8 k, i;
        u32 l, mblock, adblock, lmblock, ladblock;
        __m128i x[4*PARAL_INST_BY2];
	u8 tmp2[8], tmp3[8];
	u32 index1, index2; //, index3, index4;
	mblock = plen/8;
        lmblock = plen%8;
	adblock = adlen/8;
        ladblock = adlen%8;
       
	//Initialization: Loading Key & Nonce

	for ( i = 0; i < PARAL_INST_BY2; i++ )
	{
		x[4*i] = _mm_set_epi8(0x0, nonce[32*i+7], key[32*i+14], key[32*i+15],0x0, nonce[32*i+4],nonce[32*i+5],nonce[32*i+6],0x0, nonce[32*i+3], key[32*i+6], key[32*i+7],0x0, nonce[32*i+0],nonce[32*i+1],nonce[32*i+2]);

		x[4*i+1] = _mm_set_epi8( 0x0, nonce[32*i+23],key[32*i+30],key[32*i+31],0x0, nonce[32*i+20],nonce[32*i+21],nonce[32*i+22], 0x0, nonce[32*i+19],key[32*i+22],key[32*i+23], 0x0, nonce[32*i+16],nonce[32*i+17],nonce[32*i+18]);

		x[4*i+2] = _mm_set_epi8(0x0,key[32*i+11],key[32*i+12],key[32*i+13],0x0, key[32*i+8],key[32*i+9],key[32*i+10], 0x0,key[32*i+3],key[32*i+4],key[32*i+5],0x0, key[32*i+0],key[32*i+1],key[32*i+2]);

		x[4*i+3] = _mm_set_epi8(0x0,key[32*i+27],key[32*i+28],key[32*i+29],0x0, key[32*i+24],key[32*i+25],key[32*i+26], 0x0,key[32*i+19],key[32*i+20],key[32*i+21],0x0,key[32*i+16],key[32*i+17],key[32*i+18]);
	}
	sliscp_light192(x);
        
	for ( i = 0; i < PARAL_INST_BY2; i++ )
	{
                x[4*i+2] = x[4*i+2]^_mm_set_epi8(0x0,nonce[32*i+15],0x0,0x0, 0x0,nonce[32*i+12], nonce[32*i+13], nonce[32*i+14], 0x0,nonce[32*i+11],0x0,0x0, 0x0, nonce[32*i+8], nonce[32*i+9], nonce[32*i+10]);
                x[4*i+3] = x[4*i+3]^_mm_set_epi8(0x0, nonce[32*i+31], 0x0,0x0,0x0,nonce[32*i+28], nonce[32*i+29], nonce[32*i+30], 0x0,nonce[32*i+27], 0x0, 0x0,0x0,nonce[32*i+24], nonce[32*i+25], nonce[32*i+26]);
	}
	//Proceesing Associated Data
	for ( l = 0; l < adblock; l++ )
	{
                sliscp_light192(x);
		for ( k = 0; k < PARAL_INST_BY2; k++)
		{
			index1 = 8*l + adlen*2*k;
			index2 = 8*l + adlen*2*k+adlen;
			x[4*k+2] = x[4*k+2]^_mm_set_epi8(0x0, ad[index1+7], 0x0, 0x0, 0x0, ad[index1+4], ad[index1+5], ad[index1+6], 0x0, ad[index1+3], 0x0, 0x0, 0x0, ad[index1], ad[index1+1], ad[index1+2]);
			x[4*k+3] = x[4*k+3]^_mm_set_epi8(0x0, ad[index2+7], 0x0, 0x0, 0x0, ad[index2+4], ad[index2+5], ad[index2+6], 0x0, ad[index2+3], 0x0, 0x0, 0x0, ad[index2], ad[index2+1], ad[index2+2]);
			
			//Domain seperator
			x[4*k] = x[4*k]^_mm_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x0, 0x00);
			x[4*k+1] = x[4*k+1]^_mm_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x0, 0x00);
		}
	}
	if ( ladblock != 0 )
	{
                sliscp_light192(x);
		for ( k = 0; k < PARAL_INST_BY2; k++)
		{
			index1 = 8*l + adlen*2*k;
			index2 = 8*l + adlen*2*k+adlen;
			for ( i = 0; i < ladblock; i++ )
			{
				tmp2[i] = ad[index1+i];
				tmp3[i] = ad[index2+i];
			}
			tmp2[ladblock] = 0x80;
			tmp3[ladblock] = 0x80;
			for ( i = ladblock+1; i < 8; i++ )
			{
				tmp2[i] = 0x0;
				tmp3[i] = 0x0;
			}
			x[4*k+2] = x[4*k+2]^_mm_set_epi8(0x0, tmp2[7], 0x0, 0x0, 0x0, tmp2[4], tmp2[5], tmp2[6], 0x0, tmp2[3], 0x0, 0x0, 0x0, tmp2[0], tmp2[1], tmp2[2]);
			x[4*k+3] = x[4*k+3]^_mm_set_epi8(0x0, tmp3[7], 0x0, 0x0, 0x0, tmp3[4], tmp3[5], tmp3[6], 0x0, tmp3[3], 0x0, 0x0, 0x0, tmp3[0], tmp3[1], tmp3[2]);
			//Domain seperator
			x[4*k] = x[4*k]^_mm_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x30, 0x0, 0x00);
			x[4*k+1] = x[4*k+1]^_mm_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x30, 0x0, 0x00);
		}
	}
	//Encrypting plaintext
	for ( l = 0; l < mblock; l++ )
	{
                sliscp_light192(x);
		for ( k = 0; k < PARAL_INST_BY2; k++)
		{
			index1 = 8*l + plen*2*k;
			index2 = 8*l + plen*2*k+plen;
			plaintext[index1] = _mm_extract_epi8 (x[4*k], 2)^ciphertext[index1];
			plaintext[index1+1] = _mm_extract_epi8 (x[4*k], 1)^ciphertext[index1+1];
			plaintext[index1+2] = _mm_extract_epi8 (x[4*k], 0)^ciphertext[index1+2];
			plaintext[index1+3] = _mm_extract_epi8 (x[4*k], 6)^ciphertext[index1+3];
			plaintext[index1+4] = _mm_extract_epi8 (x[4*k], 10)^ciphertext[index1+4];
			plaintext[index1+5] = _mm_extract_epi8 (x[4*k], 9)^ciphertext[index1+5];
			plaintext[index1+6] = _mm_extract_epi8 (x[4*k], 8)^ciphertext[index1+6];
			plaintext[index1+7] = _mm_extract_epi8 (x[4*k], 14)^ciphertext[index1+7];

			x[4*k+2] = x[4*k+2]^_mm_set_epi8(0x0, plaintext[index1+7], 0x0, 0x0, 0x0, plaintext[index1+4], plaintext[index1+5], plaintext[index1+6], 0x0, plaintext[index1+3], 0x0, 0x0, 0x0, plaintext[index1], plaintext[index1+1], plaintext[index1+2]);
			
			plaintext[index2] = _mm_extract_epi8 (x[4*k+1], 2)^ciphertext[index2];
			plaintext[index2+1] = _mm_extract_epi8 (x[4*k+1], 1)^ciphertext[index2+1];
			plaintext[index2+2] = _mm_extract_epi8 (x[4*k+1], 0)^ciphertext[index2+2];
			plaintext[index2+3] = _mm_extract_epi8 (x[4*k+1], 6)^ciphertext[index2+3];
			plaintext[index2+4] = _mm_extract_epi8 (x[4*k+1], 10)^ciphertext[index2+4];
			plaintext[index2+5] = _mm_extract_epi8 (x[4*k+1], 9)^ciphertext[index2+5];
			plaintext[index2+6] = _mm_extract_epi8 (x[4*k+1], 8)^ciphertext[index2+6];
			plaintext[index2+7] = _mm_extract_epi8 (x[4*k+1], 14)^ciphertext[index2+7];
			x[4*k+3] = x[4*k+3]^_mm_set_epi8(0x0, plaintext[index2+7], 0x0, 0x0, 0x0, plaintext[index2+4], plaintext[index2+5], plaintext[index2+6], 0x0, plaintext[index2+3], 0x0, 0x0, 0x0, plaintext[index2], plaintext[index2+1], plaintext[index2+2]);
			
			//Domain seperator
			x[4*k] = x[4*k]^_mm_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x0, 0x00);
			x[4*k+1] = x[4*k+1]^_mm_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x0, 0x00);
		}
	}
	if ( lmblock != 0 )
	{
                sliscp_light192(x);
		for ( k = 0; k < PARAL_INST_BY2; k++)
		{
			index1 = 8*l + plen*2*k;
			index2 = 8*l + plen*2*k+plen;
			if ( lmblock == 1 )
			{
				plaintext[index1] = _mm_extract_epi8 (x[4*k], 2)^ciphertext[index1];
				plaintext[index2] = _mm_extract_epi8 (x[4*k+1], 2)^ciphertext[index2];
			}
			else if ( lmblock == 2 )
			{
				plaintext[index1] = _mm_extract_epi8 (x[4*k], 2)^ciphertext[index1];
				plaintext[index1+1] = _mm_extract_epi8 (x[4*k], 1)^ciphertext[index1+1];
				plaintext[index2] = _mm_extract_epi8 (x[4*k+1], 2)^ciphertext[index2];
				plaintext[index2+1] = _mm_extract_epi8 (x[4*k+1], 1)^ciphertext[index2+1];
			}
			else if ( lmblock == 3 )
			{
				plaintext[index1] = _mm_extract_epi8 (x[4*k], 2)^ciphertext[index1];
				plaintext[index1+1] = _mm_extract_epi8 (x[4*k], 1)^ciphertext[index1+1];
				plaintext[index1+2] = _mm_extract_epi8 (x[4*k], 0)^ciphertext[index1+2];
				plaintext[index2] = _mm_extract_epi8 (x[4*k+1], 2)^ciphertext[index2];
				plaintext[index2+1] = _mm_extract_epi8 (x[4*k+1], 1)^ciphertext[index2+1];
				plaintext[index2+2] = _mm_extract_epi8 (x[4*k+1], 0)^ciphertext[index2+2];
			}
			else if ( lmblock == 4 )
			{
				plaintext[index1] = _mm_extract_epi8 (x[4*k], 2)^ciphertext[index1];
				plaintext[index1+1] = _mm_extract_epi8 (x[4*k], 1)^ciphertext[index1+1];
				plaintext[index1+2] = _mm_extract_epi8 (x[4*k], 0)^ciphertext[index1+2];
				plaintext[index1+3] = _mm_extract_epi8 (x[4*k], 6)^ciphertext[index1+3];
				plaintext[index2] = _mm_extract_epi8 (x[4*k+1], 2)^ciphertext[index2];
				plaintext[index2+1] = _mm_extract_epi8 (x[4*k+1], 1)^ciphertext[index2+1];
				plaintext[index2+2] = _mm_extract_epi8 (x[4*k+1], 0)^ciphertext[index2+2];
				plaintext[index2+3] = _mm_extract_epi8 (x[4*k+1], 6)^ciphertext[index2+3];
			}
			else if ( lmblock == 5 )
			{
				plaintext[index1] = _mm_extract_epi8 (x[4*k], 2)^ciphertext[index1];
				plaintext[index1+1] = _mm_extract_epi8 (x[4*k], 1)^ciphertext[index1+1];
				plaintext[index1+2] = _mm_extract_epi8 (x[4*k], 0)^ciphertext[index1+2];
				plaintext[index1+3] = _mm_extract_epi8 (x[4*k], 6)^ciphertext[index1+3];
				plaintext[index1+4] = _mm_extract_epi8 (x[4*k], 10)^ciphertext[index1+4];
				plaintext[index2] = _mm_extract_epi8 (x[4*k+1], 2)^ciphertext[index2];
				plaintext[index2+1] = _mm_extract_epi8 (x[4*k+1], 1)^ciphertext[index2+1];
				plaintext[index2+2] = _mm_extract_epi8 (x[4*k+1], 0)^ciphertext[index2+2];
				plaintext[index2+3] = _mm_extract_epi8 (x[4*k+1], 6)^ciphertext[index2+3];
				plaintext[index2+4] = _mm_extract_epi8 (x[4*k+1], 10)^ciphertext[index2+4];
			}
			else if ( lmblock == 6 )
			{
				plaintext[index1] = _mm_extract_epi8 (x[4*k], 2)^ciphertext[index1];
				plaintext[index1+1] = _mm_extract_epi8 (x[4*k], 1)^ciphertext[index1+1];
				plaintext[index1+2] = _mm_extract_epi8 (x[4*k], 0)^ciphertext[index1+2];
				plaintext[index1+3] = _mm_extract_epi8 (x[4*k], 6)^ciphertext[index1+3];
				plaintext[index1+4] = _mm_extract_epi8 (x[4*k], 10)^ciphertext[index1+4];
				plaintext[index1+5] = _mm_extract_epi8 (x[4*k], 9)^ciphertext[index1+5];
				plaintext[index2] = _mm_extract_epi8 (x[4*k+1], 2)^ciphertext[index2];
				plaintext[index2+1] = _mm_extract_epi8 (x[4*k+1], 1)^ciphertext[index2+1];
				plaintext[index2+2] = _mm_extract_epi8 (x[4*k+1], 0)^ciphertext[index2+2];
				plaintext[index2+3] = _mm_extract_epi8 (x[4*k+1], 6)^ciphertext[index2+3];
				plaintext[index2+4] = _mm_extract_epi8 (x[4*k+1], 10)^ciphertext[index2+4];
				plaintext[index2+5] = _mm_extract_epi8 (x[4*k+1], 9)^ciphertext[index2+5];
			}
			else if ( lmblock == 7 )
			{
				plaintext[index1] = _mm_extract_epi8 (x[4*k], 2)^ciphertext[index1];
				plaintext[index1+1] = _mm_extract_epi8 (x[4*k], 1)^ciphertext[index1+1];
				plaintext[index1+2] = _mm_extract_epi8 (x[4*k], 0)^ciphertext[index1+2];
				plaintext[index1+3] = _mm_extract_epi8 (x[4*k], 6)^ciphertext[index1+3];
				plaintext[index1+4] = _mm_extract_epi8 (x[4*k], 10)^ciphertext[index1+4];
				plaintext[index1+5] = _mm_extract_epi8 (x[4*k], 9)^ciphertext[index1+5];
				plaintext[index1+6] = _mm_extract_epi8 (x[4*k], 8)^ciphertext[index1+6];
				plaintext[index2] = _mm_extract_epi8 (x[4*k+1], 2)^ciphertext[index2];
				plaintext[index2+1] = _mm_extract_epi8 (x[4*k+1], 1)^ciphertext[index2+1];
				plaintext[index2+2] = _mm_extract_epi8 (x[4*k+1], 0)^ciphertext[index2+2];
				plaintext[index2+3] = _mm_extract_epi8 (x[4*k+1], 6)^ciphertext[index2+3];
				plaintext[index2+4] = _mm_extract_epi8 (x[4*k+1], 10)^ciphertext[index2+4];
				plaintext[index2+5] = _mm_extract_epi8 (x[4*k+1], 9)^ciphertext[index2+5];
				plaintext[index2+6] = _mm_extract_epi8 (x[4*k+1], 8)^ciphertext[index2+6];
			}
			for ( i = 0; i < lmblock; i++ )
			{
				tmp2[i] = plaintext[index1+i];
				tmp3[i] = plaintext[index2+i];
			}
			tmp2[lmblock] = 0x80;
			tmp3[lmblock] = 0x80;
			for ( i = lmblock+1; i < 8; i++ )
			{
				tmp2[i] = 0x0;
				tmp3[i] = 0x0;
			}
			x[4*k+2] = x[4*k+2]^_mm_set_epi8(0x0, tmp2[7], 0x0, 0x0, 0x0, tmp2[4], tmp2[5], tmp2[6], 0x0, tmp2[3], 0x0, 0x0, 0x0, tmp2[0], tmp2[1], tmp2[2]);
			x[4*k+3] = x[4*k+3]^_mm_set_epi8(0x0, tmp3[7], 0x0, 0x0, 0x0, tmp3[4], tmp3[5], tmp3[6], 0x0, tmp3[3], 0x0, 0x0, 0x0, tmp3[0], tmp3[1], tmp3[2]);
			//Domain seperator
			x[4*k] = x[4*k]^_mm_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x50, 0x0, 0x00);
			x[4*k+1] = x[4*k+1]^_mm_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x50, 0x0, 0x00);
		}
	}
        
	//Tag generation phase
	for ( k = 0; k < PARAL_INST_BY2; k++)
	{
        	//Domain seperator
		x[4*k] = x[4*k]^_mm_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x80, 0x0, 0x00);
		x[4*k+1] = x[4*k+1]^_mm_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x80, 0x0, 0x00);
	}
        sliscp_light192(x);
	//Extracting tag
	for ( k = 0; k < PARAL_INST_BY2; k++)
	{
		_mm_storeu_si128((void *) (tag + 0 + 8*k), x[4*k+2]);
		_mm_storeu_si128((void *) (tag + 4 + 8*k), x[4*k+3]);
	}
return;
}
